from keras import layers
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFGCVit')
class SE(layers.Layer):
    def __init__(self, expansion=0.25, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.expansion = expansion

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimension of the inputs should be defined. Found `None`.')

        squeeze_filters = max(1, int(channels * self.expansion))

        # noinspection PyAttributeOutsideInit
        self.fc = [
            layers.GlobalAvgPool2D(keepdims=True),
            layers.Dense(squeeze_filters, activation='gelu', use_bias=False, name='fc/0'),
            layers.Dense(channels, activation='sigmoid', use_bias=False, name='fc/2')
        ]

        super().build(input_shape)

    def call(self, inputs, **kwargs):
        outputs = inputs
        for layer in self.fc:
            outputs = layer(outputs)

        outputs *= inputs

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({'expansion': self.expansion})

        return config



import tensorflow as tf
from keras import layers
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion
from tfgcvit.block import Block
from tfgcvit.extract import FeatExtract


@register_keras_serializable(package='TFGCVit')
class Level(layers.Layer):
    def __init__(self, depth, num_heads, window_size, keep_sizes, mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0.,
                 attn_drop=0., path_drop=0., layer_scale=None, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.depth = depth
        self.num_heads = num_heads
        self.window_size = window_size
        self.keep_sizes = keep_sizes
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.qk_scale = qk_scale
        self.drop = drop
        self.attn_drop = attn_drop
        self.path_drop = path_drop
        self.layer_scale = layer_scale

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        self.channels = input_shape[-1]
        if self.channels is None:
            raise ValueError('Channel dimension of the inputs should be defined. Found `None`.')

        path_drop = self.path_drop
        if not isinstance(self.path_drop, (list, tuple)):
            path_drop = [self.path_drop] * self.depth

        # noinspection PyAttributeOutsideInit
        self.blocks = [
            Block(window_size=self.window_size, num_heads=self.num_heads, global_query=bool(i % 2),
                  mlp_ratio=self.mlp_ratio, qkv_bias=self.qkv_bias, qk_scale=self.qk_scale, drop=self.drop,
                  attn_drop=self.attn_drop, path_drop=path_drop[i], layer_scale=self.layer_scale, name=f'blocks/{i}')
            for i in range(self.depth)]

        # noinspection PyAttributeOutsideInit
        self.global_query = [
            FeatExtract(keep_size, name=f'q_global_gen/to_q_global/{i}')
            for i, keep_size in enumerate(self.keep_sizes)]

        # noinspection PyAttributeOutsideInit
        self.resize_query = layers.Resizing(self.window_size, self.window_size)

        super().build(input_shape)

    def relative_index(self):
        offset = tf.range(self.window_size)
        offset = tf.stack(tf.meshgrid(offset, offset, indexing='ij'), axis=0)
        offset = tf.reshape(offset, [2, -1])
        offset = offset[:, :, None] - offset[:, None]

        index = offset + (self.window_size - 1)
        index = index[0] * (2 * self.window_size - 1) + index[1]
        index = tf.reshape(index, [-1])

        return index

    def call(self, inputs, *args, **kwargs):
        height, width = tf.unstack(tf.shape(inputs)[1:3])
        h_pad = (self.window_size - height % self.window_size) % self.window_size
        w_pad = (self.window_size - width % self.window_size) % self.window_size
        outputs = tf.pad(inputs, [[0, 0], [0, h_pad], [0, w_pad], [0, 0]])
        padded_height, padded_width = height + h_pad, width + w_pad

        global_query = outputs
        for layer in self.global_query:
            global_query = layer(global_query)
        global_query = self.resize_query(global_query)

        input_length = padded_height * padded_width
        query_length = self.window_size ** 2

        global_query = tf.reshape(global_query, [-1, 1, query_length, self.num_heads, self.channels // self.num_heads])
        global_query = tf.transpose(global_query, [0, 1, 3, 2, 4])
        global_query = tf.tile(global_query, [1, input_length // query_length, 1, 1, 1])
        global_query = tf.reshape(global_query, [-1, self.num_heads, query_length, self.channels // self.num_heads])

        relative_index = self.relative_index()
        for i, b in enumerate(self.blocks):
            if i % 2:
                outputs = b([outputs, relative_index, global_query])
            else:
                outputs = b([outputs, relative_index])

        outputs = outputs[:, :height, :width, ...]
        outputs.set_shape(inputs.shape)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({
            'depth': self.depth,
            'num_heads': self.num_heads,
            'window_size': self.window_size,
            'keep_sizes': self.keep_sizes,
            'mlp_ratio': self.mlp_ratio,
            'qkv_bias': self.qkv_bias,
            'qk_scale': self.qk_scale,
            'drop': self.drop,
            'attn_drop': self.attn_drop,
            'path_drop': self.path_drop,
            'layer_scale': self.layer_scale
        })

        return config



from keras import layers, models
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion
from tfgcvit.norm import LayerNorm
from tfgcvit.se import SE


@register_keras_serializable(package='TFGCVit')
class ReduceSize(layers.Layer):
    def __init__(self, keep_channels, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.keep_channels = keep_channels

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimensions of the inputs should be defined. Found `None`.')

        # noinspection PyAttributeOutsideInit
        self.norm2 = LayerNorm(name='norm2')

        # noinspection PyAttributeOutsideInit
        self.conv = [
            layers.DepthwiseConv2D(3, padding='same', activation='gelu', use_bias=False, name='conv/0'),
            SE(name='conv/2'),
            layers.Conv2D(channels, 1, use_bias=False, name='conv/3')
        ]

        # noinspection PyAttributeOutsideInit
        self.pad = layers.ZeroPadding2D(1)

        reduced_channels = channels * (2 - int(self.keep_channels))

        # noinspection PyAttributeOutsideInit
        self.reduce = layers.Conv2D(reduced_channels, 3, strides=2, use_bias=False, name='reduction')

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        residual = inputs
        for layer in self.conv:
            residual = layer(residual)

        outputs = inputs + residual
        outputs = self.pad(outputs)
        outputs = self.reduce(outputs)
        outputs = self.norm2(outputs)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        output_shape = self.pad.compute_output_shape(input_shape)
        output_shape = self.reduce.compute_output_shape(output_shape)

        return output_shape

    def get_config(self):
        config = super().get_config()
        config.update({'keep_channels': self.keep_channels})

        return config



import tensorflow as tf
from keras import backend, layers
from keras.saving.object_registration import register_keras_serializable


@register_keras_serializable(package='TFGCVit')
class SymmetricPadding(layers.ZeroPadding2D):
    def call(self, inputs):
        padding = self.padding
        data_format = self.data_format

        assert len(padding) == 2
        assert len(padding[0]) == 2
        assert len(padding[1]) == 2

        if data_format is None:
            data_format = backend.image_data_format()
        if data_format not in {'channels_first', 'channels_last'}:
            raise ValueError('Unknown data_format: ' + str(data_format))

        if data_format == 'channels_first':
            pattern = [[0, 0], [0, 0], list(padding[0]), list(padding[1])]
        else:
            pattern = [[0, 0], list(padding[0]), list(padding[1]), [0, 0]]

        return tf.pad(inputs, pattern, mode='SYMMETRIC')



import tensorflow as tf
from keras import backend, layers
from keras.utils.control_flow_util import smart_cond
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFGCVit')
class DropPath(layers.Layer):
    def __init__(self, rate, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(min_ndim=1)

        if not 0. <= rate <= 1.:
            raise ValueError(f'Invalid value {rate} received for `rate`. Expected a value between 0 and 1.')

        self.rate = rate

    def call(self, inputs, training=None, **kwargs):
        if 0. == self.rate:
            return inputs

        if training is None:
            training = backend.learning_phase()

        outputs = smart_cond(training, lambda: self.drop(inputs), lambda: tf.identity(inputs))

        return outputs

    def drop(self, inputs):
        keep = 1.0 - self.rate
        batch = tf.shape(inputs)[0]
        shape = [batch] + [1] * (inputs.shape.rank - 1)

        random = tf.random.uniform(shape, dtype=self.compute_dtype) <= keep
        random = tf.cast(random, self.compute_dtype) / keep

        outputs = inputs * random

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({'rate': self.rate})

        return config



import numpy as np
import tensorflow as tf
from keras import initializers, layers
from keras.utils.control_flow_util import smart_cond
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFGCVit')
class WindowAttention(layers.Layer):
    def __init__(self, window_size, num_heads, global_query, qkv_bias=True, qk_scale=None, attn_drop=0., proj_drop=0.,
                 **kwargs):
        super().__init__(**kwargs)
        self.input_spec = [layers.InputSpec(ndim=3), layers.InputSpec(ndim=1, dtype='int32')]
        if global_query:
            self.input_spec.append(layers.InputSpec(ndim=4, axes={1: num_heads}))

        self.window_size = window_size
        self.num_heads = num_heads
        self.global_query = global_query
        self.qkv_bias = qkv_bias
        self.qk_scale = qk_scale
        self.attn_drop = attn_drop
        self.proj_drop = proj_drop

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        self.channels = input_shape[0][-1]
        if self.channels is None:
            raise ValueError('Channel dimensions of the inputs should be defined. Found `None`.')

        self.qkv_size = 3 - int(self.global_query)

        # noinspection PyAttributeOutsideInit
        self.qkv = layers.Dense(self.channels * self.qkv_size, use_bias=self.qkv_bias, name='qkv')

        # noinspection PyAttributeOutsideInit
        self.scale = self.qk_scale or (self.channels // self.num_heads) ** -0.5

        # noinspection PyAttributeOutsideInit
        self.relative_bias = self.add_weight(
            'relative_position_bias_table',
            shape=[(2 * self.window_size - 1) ** 2, self.num_heads],
            initializer=initializers.TruncatedNormal(stddev=0.02),
            trainable=True,
            dtype=self.dtype)

        # noinspection PyAttributeOutsideInit
        self.drop_attn = layers.Dropout(self.attn_drop)

        # noinspection PyAttributeOutsideInit
        self.proj = layers.Dense(self.channels, name='proj')

        # noinspection PyAttributeOutsideInit
        self.drop_proj = layers.Dropout(self.proj_drop)

        super().build(input_shape)

    def call(self, inputs, **kwargs):
        if self.global_query:
            inputs, relative_index, q = inputs
        else:
            inputs, relative_index = inputs
        length = tf.shape(inputs)[1]

        qkv = self.qkv(inputs)
        qkv = tf.reshape(qkv, [-1, length, self.qkv_size, self.num_heads, self.channels // self.num_heads])
        qkv = tf.transpose(qkv, [2, 0, 3, 1, 4])

        if self.global_query:
            k, v = tf.unstack(qkv, 2)
        else:
            q, k, v = tf.unstack(qkv, 3)
        q *= self.scale
        attn = tf.matmul(q, k, transpose_b=True)

        bias = tf.gather(self.relative_bias, relative_index)
        bias = tf.reshape(bias, [self.window_size ** 2, self.window_size ** 2, -1])
        bias = tf.transpose(bias, perm=[2, 0, 1])
        attn = attn + bias[None]

        attn = tf.nn.softmax(attn)
        attn = self.drop_attn(attn)

        outputs = tf.transpose(tf.matmul(attn, v), perm=[0, 2, 1, 3])
        outputs = tf.reshape(outputs, [-1, length, self.channels])

        outputs = self.proj(outputs)
        outputs = self.drop_proj(outputs)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def get_config(self):
        config = super().get_config()
        config.update({
            'window_size': self.window_size,
            'num_heads': self.num_heads,
            'global_query': self.global_query,
            'qkv_bias': self.qkv_bias,
            'qk_scale': self.qk_scale,
            'attn_drop': self.attn_drop,
            'proj_drop': self.proj_drop
        })

        return config



import numpy as np
import tensorflow as tf
from keras import backend, layers, models
from keras.applications import imagenet_utils
from keras.utils import data_utils, layer_utils
from tfgcvit.embed import PatchEmbedding
from tfgcvit.norm import LayerNorm
from tfgcvit.reduce import ReduceSize
from tfgcvit.level import Level

BASE_URL = 'https://github.com/shkarupa-alex/tfgcvit/releases/download/{}/gcvit_{}.h5'
WEIGHT_URLS = {
    'gcvit_nano': BASE_URL.format('2.0.0', 'nano'),
    'gcvit_micro': BASE_URL.format('2.0.0', 'micro'),
    'gcvit_tiny': BASE_URL.format('2.0.0', 'tiny'),
    'gcvit_small': BASE_URL.format('2.0.0', 'small'),
    'gcvit_base': BASE_URL.format('2.0.0', 'base'),
    'gcvit_large': BASE_URL.format('2.0.3', 'large'),
}
WEIGHT_HASHES = {
    'gcvit_nano': '752926536d36707415c8b17d819fb1bfc48d22fd878edde1f622c76bfe23f690',
    'gcvit_micro': 'fcea210cd00d79de3fc681ddaad965ca3601077a27db256d4aacddc1154b5517',
    'gcvit_tiny': 'b55e8de5e64174619bf1ffeb11ea2d9b553ce527d6aa4370f5ade875c6e7b1f5',
    'gcvit_small': '0d9755ce464c8f4eece85493697c694ea616036d41136a602204c2fddec67b1b',
    'gcvit_base': 'bcf1dd6a59f2ef12b0aa657f30aef0ed67bb6d17e9e91186c03e4da651b28b10',
    'gcvit_large': 'ec0faee8dc7a3537d8fc64d2fdf6011cbfb468cbc37426e9596a4fdef30b475a'
}


def GCViT(
        window_size, embed_dim, depths, num_heads,
        drop_rate=0., mlp_ratio=3., qkv_bias=True, qk_scale=None, attn_drop=0., path_drop=0.1, layer_scale=None,
        model_name='gcvit', include_top=True, weights=None, input_tensor=None, input_shape=None, pooling=None,
        classes=1000, classifier_activation='softmax'):
    """Instantiates the GCViT Transformer architecture.

    Args:
      window_size: window partition size.
      embed_dim: patch embedding dimension.
      depths: depth of each level.
      num_heads: number of attention heads.
      drop_rate: dropout rate.
      mlp_ratio: ratio of mlp hidden units to embedding units.
      qkv_bias: whether to add a learnable bias to query, key, value.
      qk_scale: override default qk scale of head_dim ** -0.5 if set
      attn_drop: attention dropout rate
      path_drop: stochastic depth rate
      layer_scale: initial value for learnable block output scaling
      model_name: model name.
      include_top: whether to include the fully-connected layer at the top of the network.
      weights: one of `None` (random initialization), 'imagenet' (pre-training on ImageNet or ImageNet 21k), or the
        path to the weights file to be loaded.
      input_tensor: tensor (i.e. output of `layers.Input()`) to use as image input for the model.
      input_shape: shape tuple without batch dimension. Used to create input layer if `input_tensor` not provided.
      pooling: optional pooling mode for feature extraction when `include_top` is `False`.
        - `None` means that the output of the model will be the 3D tensor output of the last layer.
        - `avg` means that global average pooling will be applied to the output of the last layer, and thus the output
          of the model will be a 2D tensor.
        - `max` means that global max pooling will be applied.
      classes: optional number of classes to classify images into, only to be specified if `include_top` is True.
      classifier_activation: the activation function to use on the "top" layer. Ignored unless `include_top=True`.
        When loading pretrained weights, `classifier_activation` can only be `None` or `"softmax"`.

    Returns:
      A `keras.Model` instance.
    """
    if not (weights in {'imagenet', None} or tf.io.gfile.exists(weights)):
        raise ValueError('The `weights` argument should be either `None` (random initialization), `imagenet` '
                         '(pre-training on ImageNet), or the path to the weights file to be loaded.')

    if weights == 'imagenet' and include_top and classes not in {1000, 21841}:
        raise ValueError('If using `weights` as `"imagenet"` with `include_top` as true, '
                         '`classes` should be 1000 or 21841 depending on model type')

    if input_tensor is not None:
        try:
            backend.is_keras_tensor(input_tensor)
        except ValueError:
            raise ValueError(f'Expecting `input_tensor` to be a symbolic tensor instance. '
                             f'Got {input_tensor} of type {type(input_tensor)}')

    if input_tensor is not None:
        tensor_shape = backend.int_shape(input_tensor)[1:]
        if input_shape and tensor_shape != input_shape:
            raise ValueError('Shape of `input_tensor` should equals to `input_shape` if both provided.')
        else:
            input_shape = tensor_shape

    # Determine proper input shape
    input_shape = imagenet_utils.obtain_input_shape(
        input_shape,
        default_size=224,
        min_size=32,
        data_format='channel_last',
        require_flatten=False,
        weights=weights)

    if input_tensor is not None:
        if backend.is_keras_tensor(input_tensor):
            image = input_tensor
        else:
            image = layers.Input(tensor=input_tensor, shape=input_shape, dtype='float32')
    else:
        image = layers.Input(shape=input_shape)

    # Define model pipeline
    x = PatchEmbedding(embed_dim=embed_dim, name='patch_embed')(image)
    x = LayerNorm(name='patch_embed/conv_down/norm1')(x)
    x = ReduceSize(True, name='patch_embed/conv_down')(x)
    x = layers.Dropout(drop_rate, name='pos_drop')(x)

    path_drops = np.linspace(0., path_drop, sum(depths))
    keep_sizes = [
        (False, False, False),
        (False, False),
        (True,),
        (True,),
    ]

    for i in range(len(depths)):
        path_drop = path_drops[sum(depths[:i]):sum(depths[:i + 1])].tolist()
        not_last = i != len(depths) - 1

        x = Level(depth=depths[i], num_heads=num_heads[i], window_size=window_size[i], keep_sizes=keep_sizes[i],
                  mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale, drop=drop_rate, attn_drop=attn_drop,
                  path_drop=path_drop, layer_scale=layer_scale, name=f'levels/{i}')(x)
        if not_last:
            x = LayerNorm(name=f'levels/{i}/downsample/norm1')(x)
            x = ReduceSize(False, name=f'levels/{i}/downsample')(x)

    x = LayerNorm(name='norm')(x)

    if include_top or pooling in {None, 'avg'}:
        x = layers.GlobalAveragePooling2D(name='avg_pool')(x)
    elif pooling == 'max':
        x = layers.GlobalMaxPooling2D(name='max_pool')(x)
    else:
        raise ValueError(f'Expecting pooling to be one of None/avg/max. Found: {pooling}')

    imagenet_utils.validate_activation(classifier_activation, weights)
    x = layers.Dense(classes, name='head')(x)
    x = layers.Activation(classifier_activation, dtype='float32', name='pred')(x)

    # Ensure that the model takes into account any potential predecessors of `input_tensor`.
    if input_tensor is not None:
        inputs = layer_utils.get_source_inputs(input_tensor)
    else:
        inputs = image

    # Create model.
    model = models.Model(inputs, x, name=model_name)

    # Load weights.
    if 'imagenet' == weights and model_name in WEIGHT_URLS:
        weights_url = WEIGHT_URLS[model_name]
        weights_hash = WEIGHT_HASHES[model_name]
        weights_path = data_utils.get_file(origin=weights_url, file_hash=weights_hash, cache_subdir='tfgcvit')
        model.load_weights(weights_path)
    elif weights is not None:
        model.load_weights(weights)

    if include_top:
        return model

    last_layer = 'norm'
    if pooling == 'avg':
        last_layer = 'avg_pool'
    elif pooling == 'max':
        last_layer = 'max_pool'

    outputs = model.get_layer(name=last_layer).output
    model = models.Model(inputs=inputs, outputs=outputs, name=model_name)

    return model


def GCViTNano(model_name='gcvit_nano', window_size=(7, 7, 14, 7), embed_dim=64, depths=(2, 2, 6, 2),
              num_heads=(2, 4, 8, 16), path_drop=0.2, weights='imagenet', **kwargs):
    return GCViT(model_name=model_name, window_size=window_size, embed_dim=embed_dim, depths=depths,
                 num_heads=num_heads, path_drop=path_drop, weights=weights, **kwargs)


def GCViTMicro(model_name='gcvit_micro', window_size=(7, 7, 14, 7), embed_dim=64, depths=(3, 4, 6, 5),
               num_heads=(2, 4, 8, 16), path_drop=0.2, weights='imagenet', **kwargs):
    return GCViT(model_name=model_name, window_size=window_size, embed_dim=embed_dim, depths=depths,
                 num_heads=num_heads, path_drop=path_drop, weights=weights, **kwargs)


def GCViTTiny(model_name='gcvit_tiny', window_size=(7, 7, 14, 7), embed_dim=64, depths=(3, 4, 19, 5),
              num_heads=(2, 4, 8, 16), path_drop=0.2, weights='imagenet', **kwargs):
    return GCViT(model_name=model_name, window_size=window_size, embed_dim=embed_dim, depths=depths,
                 num_heads=num_heads, path_drop=path_drop, weights=weights, **kwargs)


def GCViTSmall(model_name='gcvit_small', window_size=(7, 7, 14, 7), embed_dim=96, depths=(3, 4, 19, 5),
               num_heads=(3, 6, 12, 24), mlp_ratio=2., path_drop=0.3, layer_scale=1e-5, weights='imagenet', **kwargs):
    return GCViT(model_name=model_name, window_size=window_size, embed_dim=embed_dim, depths=depths,
                 num_heads=num_heads, mlp_ratio=mlp_ratio, path_drop=path_drop, layer_scale=layer_scale,
                 weights=weights, **kwargs)


def GCViTBase(model_name='gcvit_base', window_size=(7, 7, 14, 7), embed_dim=128, depths=(3, 4, 19, 5),
              num_heads=(4, 8, 16, 32), mlp_ratio=2., path_drop=0.5, layer_scale=1e-5, weights='imagenet', **kwargs):
    return GCViT(model_name=model_name, window_size=window_size, embed_dim=embed_dim, depths=depths,
                 num_heads=num_heads, mlp_ratio=mlp_ratio, path_drop=path_drop, layer_scale=layer_scale,
                 weights=weights, **kwargs)


def GCViTLarge(model_name='gcvit_large', window_size=(7, 7, 14, 7), embed_dim=192, depths=(3, 4, 19, 5),
               num_heads=(6, 12, 24, 48), mlp_ratio=2., path_drop=0.5, layer_scale=1e-5, weights='imagenet', **kwargs):
    return GCViT(model_name=model_name, window_size=window_size, embed_dim=embed_dim, depths=depths,
                 num_heads=num_heads, mlp_ratio=mlp_ratio, path_drop=path_drop, layer_scale=layer_scale,
                 weights=weights, **kwargs)



from keras import layers
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion
from .pad import SymmetricPadding
from .se import SE


@register_keras_serializable(package='TFGCVit')
class FeatExtract(layers.Layer):
    def __init__(self, keep_size, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.keep_size = keep_size

    @shape_type_conversion
    def build(self, input_shape):
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimension of the inputs should be defined. Found `None`.')

        # noinspection PyAttributeOutsideInit
        self.conv = [
            layers.DepthwiseConv2D(3, padding='same', activation='gelu', use_bias=False, name='conv/0'),
            SE(name='conv/2'),
            layers.Conv2D(channels, 1, use_bias=False, name='conv/3'),
        ]

        if not self.keep_size:
            # noinspection PyAttributeOutsideInit
            self.pad = SymmetricPadding(1)

            # noinspection PyAttributeOutsideInit
            self.pool = layers.MaxPool2D(3, strides=2)

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        outputs = inputs
        for layer in self.conv:
            outputs = layer(outputs)

        outputs += inputs

        if not self.keep_size:
            outputs = self.pad(outputs)
            outputs = self.pool(outputs)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        if self.keep_size:
            return input_shape

        output_shape = self.pad.compute_output_shape(input_shape)
        output_shape = self.pool.compute_output_shape(output_shape)

        return output_shape

    def get_config(self):
        config = super().get_config()
        config.update({'keep_size': self.keep_size})

        return config



import tensorflow as tf
from keras import initializers, layers
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion
from tfgcvit.drop import DropPath
from tfgcvit.mlp import MLP
from tfgcvit.norm import LayerNorm
from tfgcvit.winatt import WindowAttention
from tfgcvit.window import window_partition, window_reverse


@register_keras_serializable(package='TFGCVit')
class Block(layers.Layer):
    def __init__(self, window_size, num_heads, global_query, mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0.,
                 attn_drop=0., path_drop=0., layer_scale=None, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = [layers.InputSpec(ndim=4), layers.InputSpec(ndim=1, dtype='int32')]
        if global_query:
            self.input_spec.append(layers.InputSpec(ndim=4, axes={1: num_heads}))

        self.window_size = window_size
        self.num_heads = num_heads
        self.global_query = global_query
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.qk_scale = qk_scale
        self.drop = drop
        self.attn_drop = attn_drop
        self.path_drop = path_drop
        self.layer_scale = layer_scale

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        channels = input_shape[0][-1]
        if channels is None:
            raise ValueError('Channel dimensions of the inputs should be defined. Found `None`.')

        # noinspection PyAttributeOutsideInit
        self.norm1 = LayerNorm(name='norm1')

        # noinspection PyAttributeOutsideInit
        self.attn = WindowAttention(
            window_size=self.window_size, num_heads=self.num_heads, global_query=self.global_query,
            qkv_bias=self.qkv_bias, qk_scale=self.qk_scale, attn_drop=self.attn_drop, proj_drop=self.drop, name='attn')

        # noinspection PyAttributeOutsideInit
        self.drop_path = DropPath(self.path_drop)

        # noinspection PyAttributeOutsideInit
        self.norm2 = LayerNorm(name='norm2')

        # noinspection PyAttributeOutsideInit
        self.mlp = MLP(ratio=self.mlp_ratio, dropout=self.drop, name='mlp')

        if self.layer_scale is not None:
            # noinspection PyAttributeOutsideInit
            self.gamma1 = self.add_weight(
                'gamma1',
                shape=[channels],
                initializer=initializers.Constant(self.layer_scale),
                trainable=True,
                dtype=self.dtype)

            # noinspection PyAttributeOutsideInit
            self.gamma2 = self.add_weight(
                'gamma2',
                shape=[channels],
                initializer=initializers.Constant(self.layer_scale),
                trainable=True,
                dtype=self.dtype)

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        if self.global_query:
            inputs, relative_index, query = inputs
        else:
            inputs, relative_index = inputs

        height, width = tf.unstack(tf.shape(inputs)[1:3])

        outputs = self.norm1(inputs)

        # Partition windows
        outputs = window_partition(outputs, height, width, self.window_size, self.compute_dtype)

        # W-MSA/SW-MSA
        if self.global_query:
            outputs = self.attn([outputs, relative_index, query])
        else:
            outputs = self.attn([outputs, relative_index])

        # Merge windows
        outputs = window_reverse(outputs, height, width, self.window_size, self.compute_dtype)

        # FFN
        if self.layer_scale is not None:
            outputs *= self.gamma1
        outputs = inputs + self.drop_path(outputs)
        residual = self.mlp(self.norm2(outputs))
        if self.layer_scale is not None:
            residual *= self.gamma2
        outputs += self.drop_path(residual)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def get_config(self):
        config = super().get_config()
        config.update({
            'window_size': self.window_size,
            'num_heads': self.num_heads,
            'global_query': self.global_query,
            'mlp_ratio': self.mlp_ratio,
            'qkv_bias': self.qkv_bias,
            'qk_scale': self.qk_scale,
            'drop': self.drop,
            'attn_drop': self.attn_drop,
            'path_drop': self.path_drop,
            'layer_scale': self.layer_scale
        })

        return config



from keras import layers
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFGCVit')
class PatchEmbedding(layers.Layer):
    def __init__(self, embed_dim, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.embed_dim = embed_dim

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        self.pad = layers.ZeroPadding2D(1)

        # noinspection PyAttributeOutsideInit
        self.proj = layers.Conv2D(self.embed_dim, 3, strides=2, name='proj')

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        outputs = self.pad(inputs)
        outputs = self.proj(outputs)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        output_shape = self.pad.compute_output_shape(input_shape)
        output_shape = self.proj.compute_output_shape(output_shape)

        return output_shape

    def get_config(self):
        config = super().get_config()
        config.update({'embed_dim': self.embed_dim})

        return config
